;;;; tests.lisp -- Unit tests (with Lisp-Unit) for Perl-in-Lisp

;;; Copyright 2006 Stuart Sierra

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the Lisp Lesser GNU General Public
;;; License (LLGPL) published by Franz, Inc., available at
;;; http://opensource.franz.com/preamble.html

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; Lesser GNU General Public License for more details.

(in-package :common-lisp-user)

;; (defpackage :perl-in-lisp.test
;;   (:use :common-lisp :perl-in-lisp :perl-api
;;      :lisp-unit :cffi :cffi-address)
;;   (:export #:run-tests)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (use-package :lisp-unit :perl-in-lisp))

(in-package :perl-in-lisp)
(define-test scalars-true-false
  (assert-true (pointerp (perl-newsv 0)))
  (assert-equal nil (perl-sv-true (perl-newsv 0)))
  (assert-equal t (perl-sv-true (perl-newsviv 5))))
(define-test new-scalar-integers
  (assert-equal t (perl-sv-true (perl-newsviv -256)))
  (assert-equal t (perl-sv-true (perl-newsvuv 17)))
  (assert-equal nil (perl-sv-true (perl-newsviv 0)))
  (assert-equal nil (perl-sv-true (perl-newsvuv 0)))
  (assert-equal -256 (perl-sv-2iv (perl-newsviv -256)))
  (assert-equal 17 (perl-sv-2uv (perl-newsvuv 17))))

(define-test new-scalar-floats
  (assert-equal nil (perl-sv-true (perl-newsvnv 0d0)))
  (assert-equal t (perl-sv-true (perl-newsvnv 3.1459d0)))
  (assert-equal 3.1459d0 (perl-sv-2nv (perl-newsvnv 3.1459d0))))
(define-test create-named-scalars
  (let ((x (perl-get-sv "foo" t)))
    (perl-sv-setsv-flags x (perl-newsviv 1)
                         (foreign-bitfield-value
                          'sv-flags '(:gmagic)))
    (assert-equal t (perl-sv-true
                     (perl-get-sv "foo" nil)))))
(define-test sv-setsv
  (let ((x (perl-newsv 0))
        (y (perl-newsviv 55)))
    (perl-sv-setsv-flags x y (foreign-bitfield-value
                              'sv-flags '(:gmagic)))
    (assert-equal 55 (perl-sv-2iv x))))
(define-test array-store-fetch
  (let ((a (perl-newav)))
    (perl-av-store a 3 (perl-newsviv -17))
    (assert-equal -17 (perl-sv-2iv (perl-in-lisp::av-fetch-sv a 3 nil)))))
(define-test av-clear
  (let ((a (perl-newav)))
    (perl-av-store a 0 (perl-newsviv 34))
    (perl-av-clear a)
    (assert-equal nil (perl-in-lisp::av-fetch-sv a 0 nil))))
(define-test av-push-pop
  (let ((a (perl-newav)))
    (perl-av-push a (perl-newsvpv "a" 0))
    (perl-av-push a (perl-newsvpv "b" 0))
    (assert-equal "b" (perl-in-lisp::string-from-sv (perl-av-pop a)))
    (assert-equal "a" (perl-in-lisp::string-from-sv (perl-av-pop a)))))
(define-test av-unshift
  (let ((a (perl-newav)))
    (perl-av-unshift a 10)
    (assert-equal 9 (perl-av-len a))))
(define-test av-shift
  (let ((a (perl-newav)))
    (perl-av-push a (perl-newsvpv "a" 0))
    (perl-av-push a (perl-newsvpv "b" 0))
    (assert-equal "a" (perl-in-lisp::string-from-sv (perl-av-shift a)))
    (assert-equal "b" (perl-in-lisp::string-from-sv (perl-av-shift a )))))
(define-test av-len
  (let ((a (perl-newav)))
    (perl-av-push a (perl-newsv 0))
    (perl-av-push a (perl-newsv 0))
    (perl-av-push a (perl-newsv 0))
    (assert-equal 2 (perl-av-len a))))
(define-test av-fill
  (let ((a (perl-newav)))
    (perl-av-fill a 3)
    (assert-equal 3 (perl-av-len a))))
(define-test references
  (let ((s (perl-newsv 0)))
    (assert-equality #'pointer-eq s
                     (deref-rv (perl-newrv s)))))
(define-test reference-to-array
  (let ((a (perl-newav)))
    (assert-equality #'pointer-eq a
                     (deref-rv (perl-newrv a)))))
(define-test push-and-pop-one-scalar
  (let ((x (perl-newsv 0))
        (old-stack-address *pl-stack-sp*))
    (pushs x)
    (assert-equality #'pointer-eq x
                     (address-ref *pl-stack-sp* :pointer))
    (assert-equality #'pointer-eq x (pops))
    (assert-equal old-stack-address *pl-stack-sp*)))
(define-test perl-scope
  (let ((x))
    (perl-scope
     (setf x (perl-newsviv 999))
     (perl-sv-2mortal x)
     (assert-equal 999 (perl-sv-2iv x))
     ;; still within the scope block:
     (assert-equal 1 (foreign-slot-value x 'sv 'refcnt)))
    ;; outside the scope block here:
    (assert-equal 0 (foreign-slot-value x 'sv 'refcnt))))
(define-test perl-call-pv
  (assert-equal 1  ; 1 value was return on the stack
                (progn
                  (perl-eval-pv 
"sub meaning_of_life { print \"\\nThis should be forty-two: \",
 $_[0], \"\\n\"; return $_[0]; }" t)
                  (pushmark)
                  (pushs (perl-sv-2mortal (perl-newsviv 42)))
                  (perl-call-pv "meaning_of_life"
                                (foreign-bitfield-value
                                 'perl-call-flags
                                 '(:scalar)))))
  (assert-equal 42 (perl-sv-2iv (pops))))
(define-test perl-call-function
  (assert-equal 1 (progn
                    (perl-eval-pv 
"sub meaning_of_life { print \"\\nThis should be forty-two: \",
 $_[0], \"\\n\"; return $_[0]; }" t)
                    (pushmark)
                    (pushs (perl-sv-2mortal (perl-newsviv 42)))
                    (perl-call-function "meaning_of_life" '(:scalar))))
  (assert-equal 42 (perl-sv-2iv (pops))))
(define-test eval-pv-expressions
  (assert-equal 7 (perl-sv-2iv (perl-eval-pv "3 + 4" nil)))
  (assert-equal "7" (perl-in-lisp::string-from-sv (perl-eval-pv "3 + 4" nil)))
  (assert-equal "olleh" (perl-in-lisp::string-from-sv
                         (perl-eval-pv "reverse 'hello'" nil))))
(define-test eval-pv-multi-statement
  (assert-equal "<p align=\"center\">Hello, world!</p>"
                (perl-in-lisp::string-from-sv
                 (perl-eval-pv "
package PerlInLisp::Tests;
use CGI;
my $cgi = new CGI;
$cgi->p({align=>'center'}, 'Hello, world!');" nil))))
(define-test eval-pv-local-scope
  (assert-equal 385 (perl-sv-2uv (perl-eval-pv "my $x = 385" nil)))
  (assert-equal 0 (perl-sv-2uv (perl-eval-pv "$x" nil))))
(define-test eval-pv-global-scope
  (assert-equal 200 (perl-sv-2uv (perl-eval-pv "$var = 200" nil)))
  (assert-equal 200 (perl-sv-2uv (perl-eval-pv "$var" nil))))
(define-test eval-sv
  (assert-equal 1 (perl-eval-sv (perl-newsvpv "20 + 7" 0)
                                (foreign-bitfield-value
                                 'perl-call-flags
                                 '(:scalar :eval))))
  (assert-equal 27 (perl-sv-2iv (pops))))
(define-test perl-eval-scalar
  (assert-equal 1 (perl-eval-scalar (perl-newsvpv "33+1" 0) '(:scalar)))
  (assert-equal 34 (perl-sv-2iv (pops))))
(define-test refcnts
  (let ((s (perl-newsv 0)))
    (assert-equal 1 (refcnt s))
    (assert-equal 2 (refcnt (perl-sv-newref s)))
    (perl-sv-free s)
    (assert-equal 1 (refcnt s))
    (perl-sv-free s)
    (assert-equal 0 (refcnt s))))
(define-test sv-type
  (assert-eq :null (svtype (perl-newsv 0)))
  (assert-eq :iv   (svtype (perl-newsviv 100)))
  (assert-eq :nv   (svtype (perl-newsvnv 3.14d0)))
  (assert-eq :rv   (svtype (perl-newrv (perl-newsv 0))))
  (assert-eq :pv   (svtype (perl-newsvpv "hello" 0)))
  (assert-eq :pvav (svtype (perl-newav)))
  (assert-eq :pvhv (svtype (perl-newhv))))
(define-test new-scalar-strings-of-numbers
  (assert-equal nil (perl-sv-true (perl-newsvpv "0" 0)))
  (assert-equal "0" (string-from-sv (perl-newsviv 0)))
  (assert-equal "-256" (string-from-sv (perl-newsviv -256)))
  (assert-equal "3.14" (string-from-sv (perl-newsvnv 3.14d0))))

(define-test new-scalar-strings-to-booleans
  (assert-equal t (perl-sv-true (perl-newsvpv "foo" 0)))
  (assert-equal t (perl-sv-true (perl-newsvpvn "foo" 3)))
  (assert-equal t (perl-sv-true (perl-newsvpv "1" 0)))
  (assert-equal nil (perl-sv-true (perl-newsvpv "" 0)))
  (assert-equal nil (perl-sv-true (perl-newsvpv "0" 0))))

(define-test new-scalar-strings
  (assert-true (pointerp (perl-newsvpv "hello" 0)))
  (assert-equal "hello" (string-from-sv (perl-newsvpv "hello" 0)))
  (assert-equal "good" (string-from-sv (perl-newsvpvn "goodbye" 4))))
(define-test string-containing-null
  (assert-equal (let ((string (copy-seq "abcde")))
                  (setf (aref string 2) (code-char 0))
                  string) ; "ab" + NULL + "de"
                (string-from-sv (perl-eval-pv "qq{ab\\00de}" 0))))
(define-test sv-set-iv-uv-nv-pv
  (let ((x (perl-newsv 0)))
    (perl-sv-setiv-mg x -256)
    (assert-equal -256 (perl-sv-2iv x))
    (perl-sv-setuv-mg x 88)
    (assert-equal 88 (perl-sv-2uv x))
    (perl-sv-setnv-mg x 3.1459d0)
    (assert-equal 3.1459d0 (perl-sv-2nv x))
    (perl-sv-setpv-mg x "hello" 0)
    (assert-equal "hello" (string-from-sv x))
    (perl-sv-setpvn-mg x "goodbye" 4)
    (assert-equal "good" (string-from-sv x))))
(define-test get-scalar-by-name
  (let ((x (get-scalar-by-name "nines")))
    (declare (ignore x))
    (perl-eval-pv "$nines = 999;" nil)
    (assert-equal
     999
     (perl-sv-2uv
      (get-scalar-by-name "nines")))))
(define-test stack
  (let ((things (list 1 2 "hello" -27)))
    (perl-scope
     (push-mortals-on-stack things)
     (assert-equal things (get-from-stack (length things))))))
(define-test call-perl
  (eval-perl "use CGI;")
  (assert-equal "<p>Hello, 1999</p>"
                (call-perl "CGI::p" :string nil "Hello," 1999)))
(define-test eval-perl
  (assert-equal 7 (eval-perl "3+4"))
  (assert-equal "abcdef" (eval-perl "'abc' . 'def'"))
  (assert-equal '(1 2 3) (eval-perl "[1, 2, 3];")))

(define-test eval-perl-with-hash
  (let ((hash
         (eval-perl
          "{aa => 1, bb => 3.14, cc => 'hello'};")))
    (assert-true (hash-table-p hash))
    (assert-equal 1 (gethash "aa" hash))
    (assert-true (< 3.13 (gethash "bb" hash) 3.15))
    (assert-equal "hello" (gethash "cc" hash))
    ))

(define-test eval-perl-creating-hash
  (let ((hash (make-hash-table)))
    (setf (gethash 'key1 hash) "one")
    (setf (gethash 'key2 hash) "two")
    (setf (gethash 'key3 hash) "three")
    (let ((new-hash (lisp-from-perl (perl-from-lisp hash))))
      (assert-true (hash-table-p new-hash))
      (assert-equal "one" (gethash "KEY1" new-hash))
      (assert-equal "two" (gethash "KEY2" new-hash))
      (assert-equal "three" (gethash "KEY3" new-hash)))))
(define-test lisp-from-perl-scalars
  (assert-equal 42 (lisp-from-perl (perl-from-lisp 42)))
  (assert-equal "Hello, world!"
                (lisp-from-perl (perl-from-lisp "Hello, world!")))
  (assert-true
   ;; we can't get exact equality from floats
   (< 3.14589 (lisp-from-perl (perl-from-lisp 3.1459)) 3.14591))
  (assert-equal nil (lisp-from-perl (perl-from-lisp nil))))
